# -*- python -*-

# Do things with processes.

import subprocess
from .alerts import *

# we take these characters as the indication to run a command in the shell, if
# not otherwise indicated
shellmeta = "\"'`|&;[(<>)]*?$"

def backquote(command, shell=None, full_result=False, silent=False):
    """Similar to Perl's `command` feature: run process, return result.

    If command is a tuple or a list, run it directly. Otherwise, make it a
    string if necessary and:
    
        If shell is True, run command as shell command line with "/bin/sh".

        If shell is otherwise true, use it as the shell and run command in it.
    
        If shell is None (or unspecified), run command with "/bin/sh" if it
        contains shell meta characters. Otherwise, split the string into a list
        and run it directly.

        If shell is otherwise false, split the string into a list and run it
        directly.

    If full_result is false (the default), return only stdout as a string. In
    this case, a ChildProcessError is raised if the exit status of the command
    is non-zero or stderr is not empty. This can be suppressed by setting silent
    to true.

    If full_result is true, return a tuple of (stdout, stderr, exit status). No
    exception for exit status or stderr is raised, regardless of the value of
    silent.

    In any case, however, there will be an exception raised if the
    called program cannot be found.

    """
    run_shell = False                   # for testing
    if not isinstance(command, (list, tuple)):
        command = str(command)
        if shell:
            if shell is True:
                shell = "/bin/sh"
            command = [shell, "-c", command]
            run_shell = True
        elif shell is None:
            if any([ ch in shellmeta for ch in command ]):
                command = ["/bin/sh", "-c", command]
                run_shell = True
            else:
                command = command.split()
        else:
            command = command.split()

    with subprocess.Popen(command, stdin=subprocess.DEVNULL,
                          stderr=subprocess.PIPE,
                          stdout=subprocess.PIPE) as proc:
        proc.wait()
        result = (proc.stdout.read().decode("utf-8"),
                  proc.stderr.read().decode("utf-8"),
                  proc.returncode)
    if full_result:
        if full_result == "plus":
            return result, run_shell
        return result
    else:
        if not silent and (result[1] or result[2]):
            raise ChildProcessError("command {} exited status {}; stderr: {}"
                                    .format(command, result[2],
                                            repr(result[1])))
        return result[0]



# Copyright (C) 2020 Juergen Nickelsen <ni@w21.org>, see LICENSE.

"""POSIX-conformant command-line option parser (plus long options)
See the parse() function for details. Build $__package_version$
"""

import os
import sys
import copy

# Exception argument string and additional value (used to generate README)
ErrorNotopt = "unknown option"                     # option
ErrorArg = "option does not take an argument"      # option
ErrorNoarg = "option needs argument"               # option
ErrorIntarg = "option argument must be integer"    # option
ErrorMinarg = "too few arguments, needs at least"  # minimum
ErrorMaxarg = "too many arguments, at most"        # maximum

class OptionError(Exception):
    pass

class OptionValueContainer:
    def __init__(self, descriptors, args):
        """Arguments: dict optchar => descriptor; command-line args.

        See the parse() function below for details.

        """
        self._opts = copy.deepcopy(descriptors)
        _keywords = ("_arguments", "_help_header", "_help_footer",
                 "_usage", "_program")
        for opt, desc in self._opts.items():
            if opt.startswith("_"):
                assert opt in _keywords, "keyword unknown: " + repr(opt)
            # While the following is grossly redundant, it is a kludge to
            # finally reach 100% test coverage; apparently unpossible if both
            # statements are under the same "if". Good luck we don't need to
            # optimise for speed. :-(
            if opt.startswith("_"):
                continue
            assert type(opt) == str and len(opt) == 1, \
              "option key must be string of length 1: " + repr(opt)
            assert type(desc) == tuple and len(desc) in (4, 5), \
              "descriptor not sequence len 4 or 5: -" + opt

            name, typ, default, *_ = desc
            assert isinstance(name, str), "option name is not a string: -" + opt
            if type(typ) == type:
                assert typ in (bool, int, str), "invalid option type: -" + opt
            else:
                assert callable(typ), "invalid option type: -"+opt
            self.__dict__[name] = default

        if "?" not in self._opts:
            self._opts["?"] = \
                ("help", self.ovc_help, None, "show help on options and things")
        if "h" not in self._opts:
            self._opts["h"] = self._opts["?"]
        for field in _keywords:
            self.__dict__[field] = self._opts.get(field)
        if not self._program:
            self._program = os.path.basename(sys.argv[0])
        self._long = { v[0].replace("_", "-"): v
                       for k, v in self._opts.items() if len(k) == 1 }
        self._args = args[:]
        self._min = self._max = None
        if type(self._arguments) == list:
            min = max = 0
            inf = False
            for arg in self._arguments:
                if "..." in arg:
                    inf = True
                if arg.startswith("["):
                    max += len(arg.split(" "))
                elif not arg == "...":
                    min += 1
                    max += 1
            self._min = min
            self._max = None if inf else max
            self._arguments = " ".join(self._arguments)


    def _parse(self):
        while self._args and self._args[0].startswith("-"):
            arg = self._args.pop(0)
            if arg == "--": break
            if arg.startswith("--"):
                self._have_opt(arg[2:])
            else:
                arg = arg[1:]
                while arg:
                    arg = self._have_opt(arg[0], arg[1:])
        if self._min is not None and len(self._args) < self._min:
            raise OptionError(ErrorMinarg, self._min)
        if self._max is not None and len(self._args) > self._max:
                raise OptionError(ErrorMaxarg, self._max)


    def _have_opt(self, opt, arg=None):
        value = None
        if len(opt) > 1:
            parts = opt.split("=", 1)
            if len(parts) > 1:
                opt, value = parts
            desc = self._long.get(opt)
        else:
            desc = self._opts.get(opt)
        if not desc:
            raise OptionError(ErrorNotopt, opt)
        name, typ, defval, *_ = desc
        if typ == bool:
            if value:
                raise OptionError(ErrorArg, opt)
            self.__dict__[name] += 1
        else:
            if typ not in (str, int):
                value = typ()
            elif arg:
                value = arg
                arg = ""
            self._set_optarg(opt, desc, value)
        return arg


    def _set_optarg(self, opt, desc, value):
        if value is None:
            if not self._args:
                raise OptionError(ErrorNoarg, opt)
            value = self._args.pop(0)
        if desc[1] == int:
            try:
                value = int(value)
            except:
                raise OptionError(ErrorIntarg, opt)
        if isinstance(getattr(self, desc[0], None), list):
            getattr(self, desc[0]).append(value)
        else:
            setattr(self, desc[0], value)


    def ovc_help(self):
        """Print the help message and exit."""
        print(self.ovc_help_msg())
        sys.exit()

        
    def ovc_help_msg(self):
        """Return a detailed help message."""
        msg = self.ovc_usage_msg() + "\n"
        if self._help_header:
            msg += self._help_header + "\n\n"
        for opt in sorted(self._opts.keys()):
            if opt.startswith("_"):
                continue
            desc = self._opts[opt]
            arg = ""
            if desc[1] in (str, int):
                arg = " " + (desc[4] if len(desc) == 5 else "ARG")
            msg += " -%s, --%s%s:\n    %s" % (
                opt, desc[0].replace('_', '-'), arg, desc[3])
            if desc[1] in (int, str):
                msg += " (%s arg, default %s)" % (
                    desc[1].__name__, repr(desc[2]))
            msg += "\n"
        if self._help_footer:
            msg += "\n" + self._help_footer
        return msg


    def ovc_usage(self, error="", exit_status=64):
        """Print usage message (with optional error message) and exit."""
        out = sys.stdout if not exit_status else sys.stderr
        if error:
            print(self._program + ":", error, file=out, end="\n\n")
        print(self.ovc_usage_msg(), file=out)
        print("use '-?' option to get more help", file=out)
        sys.exit(exit_status)


    def ovc_usage_msg(self):
        """Return a brief usage message."""
        args = ""
        if self._arguments is None:
            args = " <arguments>"
        elif self._arguments:
            args = " " + self._arguments
        noarg = ""
        w_arg = []
        for key, desc in self._opts.items():
            if len(key) > 1 or key in "h?":
                continue
            if desc[1] is str:
                w_arg.append((key, (desc[4] if len(desc) == 5 else "ARG")))
            else:
                noarg += key
        options = " "
        if noarg:
            options += "[-" + "".join(sorted(noarg)) + "]"
        for opt in w_arg:
            options += " [-{} {}]".format(opt[0], opt[1])
        return self._usage or "usage: " + self._program + options + args

    def ovc_values(self):
        """Return a dict of options and their values (for testing)."""
        return { key: val for key, val in self.__dict__.items()
                 if not key.startswith("_") }


def parse(descriptors, args=sys.argv[1:], exit_on_error=True):
    """Parse the command line options according to the specified descriptors.

    Keys of the descriptors dictionary are options or keywords. In case
    of an option, the key is the single option character, and the value
    is a tuple of four or five fields:

      (1) name of the option, used in the returned namespace and as the
      name of the corresponding long option name (after replacing
      underscores with dashes)

      (2) type of the option, which may be bool for options without
      arguments (actually counters), or str or int for options with an
      argument of the respective type

      (3) default value, which can be a starting counter (or False) for
      bool options, or an integer or string value for int or str
      options, respectively, or a list, to which each option argument
      will be appended (for multi-value options)

      (4) description of the option for the help text

      (5) (optional) name of the option's argument for the help text
      (defaults to 'ARG')

    A key may also be one of these keywords:

      "_arguments": string to print in the usage to describe the
      non-option arguments, or, for argument count checking, a sequence
      with the argument names:
    
         - a normal string counts as one argument towards minimum and
           maximum

         - if it contains '...', there is no maximum the number of
           arguments

         - if it begins with '[', it is optional; if it can be split by
           blanks into multiple words, each one counts toward the
           maximum; e.g. "[param1 param2 param3]" increases the maximum
           by 3, but not the minimum

      "_help_footer": string to print with 'help' after the option
      explanations

      "_help_header": string to print with 'help' before the option
      explanations

      "_program": string to use as program name for help and usage
      message instead of sys.argv[0]

      "_usage": string to use as usage message instead of the default
      constructed one

    If no '?' or 'h' option is specified, they will default to a long
    form of '--help' and a 'help' function, which will be called
    immediately when the option is seen. It prints a brief summary of
    the program's parameters and a description of the options, framed
    by the _help_header and the _help_footer; it terminates the program
    after printing the message.

    In case of a normal return of parse() (i.e. options and number of
    arguments okay), it returns an OptionValueContainer and a list of
    the remaining command line arguments. Example:

      ovc, args = pgetopt.parse({
      # opt: (name,          type, default value, helptext[, arg name])
        "s": ("schmooze",    bool, 0,    "more schmooziness"),
        "o": ("output_file", str,  None, "output file (or stdout)", "NAME"),
        "n": ("repetitions", int,  3,    "number of repetitions"),
        "d": ("debug",       str, [],    "debug topics", "DEBUG_TOPIC"),
      # keyword:        value
        "_arguments":   ["string_to_print", "..."],
        "_help_header": "print a string a number of times",
        "_help_footer": "This is just an example program.",
      }

      On return, ovc has the following fields:
        ovc.schmooze:    number of -s options counted,
        ovc.output_file: parameter of -o or --output-file, or None
        ovc.repetitions: parameter of -n or --repetitions, or 3
        ovc.debug:       list with all parameters given to -d or --debug

    Parameters to int or str options are taken from the next argument;
    with long options, "--option=parameter" is also possible.

    Other potentially useful fields of ovc:
      ovc.ovc_help():  help function
      ovc.ovc_usage(): usage function
    
      ovc.ovc_help_msg(),
      ovc.ovc_usage_msg(): get corresponding messages as strings

    """
    ovc = OptionValueContainer(descriptors, args)
    try:
        ovc._parse()
        return ovc, ovc._args
    except OptionError as e:
        if exit_on_error:
            ovc.ovc_usage(e.args[0] + ": " + repr(e.args[1]))
        raise(e)

# EOF



# -*- fill-column: 72 -*-
"""Do sane SIGINT and SIGPIPE handling in a decorator.

Other than is traditional (and reasonable) in Unix programs, Python
makes a big fuss when it catches a SIGINT or SIGPIPE. That includes a
stacktrace and ugly error messages for totally normal conditions, in
which the apppropriate reaction is nothing but "shut TF up and exit".

But not only is Python's default handler for these signals so
over-excited; even if you catch the exception and end the program, an
unwanted message to stderr still crops up. You have to suppress that by
explicitly closing sys.stderr -- only then Python shuts TF up.

Stupid, stupid Python! This is the more aggravating as Python usually
makes all the right choices, much more so than many other languages.
*sigh*

This decorator repairs that. Wrap it around your main() and be set.

"""

def sanesighandler(func):
    """Decorator: exit the program silently on SIGINT and SIGPIPE."""
    import sys
    
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except KeyboardInterrupt:
            # get the prompt nicely on a fresh new line, not after a ^C
            print(file=sys.stderr)
            sys.stderr.close()
            sys.exit(130)
        except BrokenPipeError:
            sys.stderr.close()
            sys.exit(141)
    return wrapper



# assorted smallish functions

import collections

# Yes, this is a bit silly. But hey...
means_true = \
  set("yes y sure ja j jou si on oui t true  aye 1 affirmative".split())
means_false = \
  set("no n  nope nein nee   off non f false nay 0 negative".split())

def boolish(value, default=None):
    """Return a truth value for the argument.

    If that cannot be determined, fall back to default (if not None) or raise a
    ValueError exception. This can be used for parsing config files (that aren't
    Python) or interactive answers or the like.

    """

    val = value.strip().lower()
    if val in means_true:
        return True
    if val in means_false:
        return False
    if default is None:
        raise ValueError("value '{}' cannot be understood as false or true".
                         format(value))
    else:
        return default


def flatten(seq):
    """Flatten a nested sequence into a flat one with the same elements.

    Return a flat generator object containing just the elements. If the
    argument is a string or not a sequence, the generator object will
    contain just the argument.

    """
    if not isinstance(seq, str) and isinstance(seq, collections.abc.Iterable):
        for elem in seq:
            yield from flatten(elem)
    else:
        yield seq


def maybe_int(arg):
    """Return the corresponding int if the arguments represents one, or None."""
    try:
        return int(arg)
    except:
        return None


def is_int(arg):
    """Return True if the arguments represents an int, or False.
    
    The argument may be not an int (maybe e.g. a string), but if it
    can be cast to an int, it represents an int.

    """
    return maybe_int(arg) is not None

def is_sequence(arg):
    """Return True iff the argument is a sequence other than string."""
    if isinstance(arg, (str, collections.UserString)):
        return False
    return isinstance(arg, collections.abc.Sequence)



import fcntl
import struct
import termios

from .alerts import debug

# Here used to be terminal_size(). In between I have learned of
# os.get_terminal_size(), which has the exact same call signature.

ttyo_file = None
ttyi_file = None

def ttyo(close=False):
    global ttyo_file
    if close:
        if ttyo_file:
            ttyo_file.close()
            ttyo_file = None
        return
    if not ttyo_file:
        ttyo_file = open("/dev/tty", "w")
    return ttyo_file

def ttyi(close=False):
    global ttyi_file
    if close:
        if ttyi_file:
            ttyi_file.close()
            ttyi_file = None
        return
    if not ttyi_file:
        ttyi_file = open("/dev/tty")
    return ttyi_file

def ptty(*args, **kwargs):
    print(*args, file=ttyo(), **kwargs)

# EOF



# Config class with configuration file reader

import errno

from .namespace import Namespace
from .kvs import parse_kvs
# from .alerts import debug

class Config(Namespace):
    """Name space class used to build a config object."""

    def update(self, new_values, reject_unknown=True):
        super().update(new_values, skip_underscore=True,
                       reject_unknown=reject_unknown)

    def set(self, key, value, reject_unknown=True):
        if key not in self.__dict__ and reject_unknown:
            raise KeyError("variable not in config: " + repr(key))
        self.__dict__[key] = value

    def get(self, key):
        if key not in self.__dict__:
            raise KeyError("variable not in config: " + repr(key))
        return self.__dict__[key]

    def load_from(self, filename, reject_unknown=True, file_must_exist=True):
        """Read a configuration from file 'filename'."""
        try:
            with open(filename, "r") as f:
                contents = f.read()
        except FileNotFoundError as exc:
            if not file_must_exist:
                return None
            else:
                raise exc
        new_locals = {}
        try:
            exec(contents, globals(), new_locals)
        except Exception as e:
            raise type(e)("Error in config file: {}; {}".format(
                filename, e
            ))
        self.update(new_locals, reject_unknown=reject_unknown)
        return True

    def load_config_files(self, config_files, notice_func=None,
                          reject_unknown=True, files_must_exist=False):
        """Read the configuration from the config files.

        Optional "notice_func" is a function to print a message
        about a config file being loaded.

        """
        loaded = 0
        for file in config_files:
            if self.load_from(file, reject_unknown=reject_unknown,
                              file_must_exist=files_must_exist):
                loaded += 1
                if notice_func:
                    notice_func("configuration loaded from", file)
        return loaded

    def update_from_string(self, cfgstring, reject_unknown=True, intvals=True):
        """Update the configuration from a key-value string.

        This can be used to pass config snippets on the command line.
        The string can look like e.g. this:

        "foo=bar,dang=[1,2,15],d={a=b,c=[d,e,f],quux=blech},e=not"

        """
        self.update(parse_kvs(cfgstring, intvals=intvals),
                    reject_unknown=reject_unknown)



# table generator

import jpylib as y
import functools


class Table:

    def __init__(self, *, corner=["", "", "", ""], border=["", "", "", ""],
                 hsep=["", ""], vsep=["", ""], tb_cross=["", ""],
                 lb_cross=["", ""], rb_cross=["", ""], bb_cross=["", ""],
                 hl_cross=["", ""], nl_cross=["", ""], cell_pad=[1, 1],
                 pad_char=" ", template=None, align=None, data=None):
        """Initialise a Table formatting parameter set.
        
        Arguments:
        * corner[4]:   top left, top right, bottom left, bottom right corners
        * border[4]:   top, left, right, bottom border sans crossing or corner
        * hsep[2]:     horizontal separator, after first column and others
        * vsep[2]:     vertical separator, after first row and others
        * tb_cross[2]: top border crossing, first and others
        * lb_cross[2]: left border crossing, first and others
        * rb_cross[2]: right border crossing, first and others
        * bb_cross[2]: bottom border crossing, first and others
        * hl_cross[2]: header separator line crossing, first and others
        * nl_cross[2]: normal separator line crossing, first and others
        * cell_pad[2]: minimum cell padding, left and right (integers)
        * pad_char:    padding character
        * template:    a template 7 x 7 drawing describing the table
        * align:       alignment descriptor string, 1 char per column, l/r/c;
                       may be sequence of 2 for first and following rows;
                       asterisk at the end means default for folloing columns
                       is the character in front of the asterisk
        """
        self.__dict__.update(locals())
        if template:
            self._from_template(template)

        if cell_pad is None:
            self.cell_pad = [0, 0]
        elif y.is_sequence(cell_pad):
            if not all(map(y.is_int, cell_pad)):
                raise ValueError("cell_pad is not a sequence of int, but {}"
                                 .format(repr(cell_pad)))
            if len(cell_pad) == 1:
                self.cell_pad = cell_pad * 2
            elif len(cell_pad) != 2:
                raise ValueError("cell_pad is not a sequence len 1 or 2, but {}"
                                 .format(len(cell_pad)))
        elif y.is_int(cell_pad):
            self.cell_pad = [cell_pad, cell_pad]
        else:
            raise ValueError("cell_pad is not None or int or seq of 2 ints: {}"
                             .format(repr(cell_pad)))

        # Always have a separate alignment for the first and the following
        # lines. They need not be different, though.
        self.defaultalign = [None, None]
        if align is None:
            self.align = ["", ""]
        elif type(align) == str:
            self.align = [align, align]
        elif y.is_sequence(align):
            if not all([isinstance(elem, (str, type(None))) for elem in align]):
                raise ValueError("align is not a sequence of str|None, but {}"
                                 .format(repr(align)))
            elif len(align) == 1:
                self.align = [align[0], align[0]]
            elif len(align) == 2:
                self.align = [align[0], align[1]]
            else:
                raise ValueError("align must be sequence of len 1 or 2, not {}"
                                 .format(len(align)))
        else:
            raise ValueError("align must be str or None or a seq of str|None, "
                             + "but is {}".format(repr(align)))
        # Set default alignment if align ends with "*"
        for i in (0, 1):
            if self.align[i]:
                if self.align[i].endswith("*") and len(self.align[i]) > 1:
                    self.defaultalign[i] = self.align[i][-2]
                    self.align[i] = self.align[i][:-1]
        if data:
            self._fill_table(data)

    def _from_template(self, template):
        def s(char):
            return "" if char == "0" else char

        # If a line ends with blanks, this can confuse the reader, as well as
        # the author and maybe even a text editor. To avoid that, a template
        # line may end with a guard character ";", which is then stripped from
        # the line.
        tl = [l.rstrip(";") for l in template.split("\n") if l]
        if len(tl) != 7:
            raise ValueError("template must have 7 lines, not {}"
                             .format(len(tl)))
        for i, l in enumerate(tl):
            if len(l) != 7:
                raise ValueError("template line is not 7 chars: {} (line {})"
                                 .format(len(l), i))
        self.corner = [s(tl[0][0]), s(tl[0][6]), s(tl[6][0]), s(tl[6][6])]
        self.border = [s(tl[0][1]), s(tl[1][0]), s(tl[1][6]), s(tl[6][1])]
        self.hsep = [s(tl[1][2]), s(tl[1][4])]
        self.vsep = [s(tl[2][1]), s(tl[4][1])]
        self.tb_cross = [s(tl[0][2]), s(tl[0][4])]
        self.lb_cross = [s(tl[2][0]), s(tl[4][0])]
        self.rb_cross = [s(tl[2][6]), s(tl[4][6])]
        self.bb_cross = [s(tl[6][2]), s(tl[6][4])]
        self.hl_cross = [s(tl[2][2]), s(tl[2][4])]
        self.nl_cross = [s(tl[4][2]), s(tl[4][4])]
        

    def _fill_table(self, data):
        """Assess and store the table data."""
        self.data = data
        self.cols = 0                   # maximum column number
        self.rows = 0                   # maximum row number
        self.col_width = []             # maximum item width per columns

        for row, data_line in enumerate(data):
            self.rows = row + 1
            for col, data_item in enumerate(data_line):
                col_width = self.col_width
                if col >= self.cols:
                    self.cols = col + 1
                    self.col_width.append(0)
                self.col_width[col] = \
                    max(self.col_width[col], len(str(data_item)))
        return self                     # so we can do Table().from().string()

    def _padded_item(self, item, width, alignment):
        """Return a string of the item with left and right padding."""
        padding = width - len(item)
        if alignment is None:
            if y.is_int(item):
                alignment = "r"
            else:
                alignment = "l"                
        if alignment == "l":
            lpad = 0
            rpad = padding
        elif alignment == "r":
            lpad = padding
            rpad = 0
        elif alignment == "c":
            lpad = int(padding / 2)
            rpad = padding - lpad
        else:
            raise ValueError("invalid char in alignment: {}"
                             .format(repr(alignment)))
        return (self.pad_char * (lpad + self.cell_pad[0])
                + str(item)
                + self.pad_char * (self.cell_pad[1] + rpad))

    def _vert_sep(self, left_border, right_border, line, cross1, cross2):
        r = []
        if line:
            r.append(left_border)
            had_first_col = False
            cross = cross1
            for col in range(self.cols):
                if had_first_col:
                    r.append(cross)
                    cross = cross2
                had_first_col = True
                r.append(line * (self.col_width[col] + sum(self.cell_pad)))
            r.append(right_border)
        result = "".join(r)
        if result:
            return result + "\n"
        return ""
            

    def _alignment(self, row, column):
        index = 0 if row == 0 else 1
        align = self.align[index]
        if column >= len(align or ""):
            return self.defaultalign[index]
        return align[column]


    def format(self, data=None):
        """Return the formatted Table as a string. Data must be present."""

        if data:
            self._fill_table(data)
        assert self.data, "Table has no data yet, so cannot be formatted."
        r = [self._vert_sep(self.corner[0], self.corner[1],
                            self.border[0], self.tb_cross[0], self.tb_cross[1])]
        had_first_row = False
        left_cross, int_cross1, int_cross2, right_cross = (
            self.lb_cross[0], self.hl_cross[0],
            self.hl_cross[1], self.rb_cross[0]
        )
        vsep = self.vsep[0]
        for row, data_line in enumerate(self.data):
            if had_first_row:
                r.append(self._vert_sep(left_cross, right_cross,
                                        vsep, int_cross1, int_cross2))
                left_cross, int_cross1, int_cross2, right_cross = (
                    self.lb_cross[1], self.nl_cross[0],
                    self.nl_cross[1], self.rb_cross[1]
                )
                vsep = self.vsep[1]
            had_first_row = True

            had_first_col = False
            hsep = self.hsep[0]
            r.append(self.border[1])
            for col, data_item in enumerate(data_line):
                if had_first_col:
                    r.append(hsep)
                    hsep = self.hsep[1]
                had_first_col = True
                r.append(self._padded_item(str(data_item), self.col_width[col],
                                           self._alignment(row, col)))
            r.append(self.border[2])
            r.append("\n")
                
        # bottom border
        r.extend(self._vert_sep(self.corner[2], self.corner[3],
                                self.border[3], *self.bb_cross))
        return "".join(r)



#!/usr/bin/env python3
# Have all support function modules here at hand.

import os
import pwd
import sys

from .pgetopt import parse as pgetopts
from .alerts import L_ERROR, L_NOTICE, L_INFO, L_DEBUG, L_TRACE, \
    alert_config, alert_level, alert_level_name, \
    alert_level_up, alert_level_zero, is_notice, is_info, is_debug, is_trace, \
    debug_vars, fatal, err, notice, info, debug, trace
from .fntrace import tracefn
from .stringreader import StringReader
from .kvs import parse_kvs
from .namespace import Namespace
from .config import Config
from .secrets import putsecret, getsecret, getsecret_main, putsecret_main
from .sighandler import sanesighandler
from .terminal import ttyi, ttyo, ptty
from .capture import outputCaptured, outputAndExitCaptured
from .process import backquote
from .assorted import boolish, flatten, maybe_int, is_int, is_sequence
from .time import isotime, iso_time, iso_time_ms, iso_time_us
from .table import Table

version = "2020.901.1046"
program = os.path.basename(sys.argv[0])
real_home = pwd.getpwuid(os.getuid()).pw_dir
home = os.environ.get("HOME") or real_home
verbosity_option = ("verbose", alert_level_up, alert_level(L_NOTICE),
                    "increase verbosity")



# print alerts and other messages depending on a verbosity level

import os
import sys
import syslog
import inspect
import jpylib as y
from .config import Config

# properties of the alert levels; the decoration will be formatted with the
# locals() values
alert_levels = (
    # level name, message decoration, fd (will look up later to make output
    # capturing work)
    ("L_ERROR", "{cfg.program}: Error:", 2),
    ("L_NOTICE", None,                     2),
    ("L_INFO",   None,                     2),
    ("L_DEBUG",  "DBG",                    2),
    ("L_TRACE",  "TRC",                    2),
)
for i, props in enumerate(alert_levels):
    name, *_ = props
    locals()[name] = i

# the module configuration; will be initialised in alert_init()
cfg = None

def alert_config(*, decoration=None, fd=None, level=None, program=None,
                 syslog_facility=None, syslog_prio=None, reset_defaults=None,
                 timestamps=None):
    """Reset everything to the specified or default values."""
    global cfg
    if not any(locals().values()) or reset_defaults:
        cfg = Config(
            # decoration to print before a message, per level
            decoration=[level[1] for level in alert_levels],
            
            # program name to use in a message
            program=os.path.basename(sys.argv[0]),
            
            # syslog facility; if set, syslog will be used
            syslog_facility=None,

            # syslog priority
            syslog_prio = [
                syslog.LOG_ERR,
                syslog.LOG_NOTICE,
                syslog.LOG_INFO,
                syslog.LOG_DEBUG,
                None,                   # don't let this go to syslog
            ],
            
            # status: syslog has been opened
            syslog_opened=False,
            
            # fd to print message to, per level
            fd=[level[2] for level in alert_levels],

            # current alert level
            level=L_NOTICE,
            
            # maximum alert level
            max_level=len(alert_levels)-1,

            # print timestamps with messages
            timestamps=False,

            # had any errors yet?
            had_errors=False,
        )
    del reset_defaults
    for var, value in locals().items():
        if value is not None:
            cfg.set(var, value)
    if cfg.timestamps is True:
        cfg.timestamps = y.isotime

def alert_init(**kwargs):
    alert_config(reset_defaults=True, **kwargs)

alert_init()


def alert_redirect(level, file):
    """Redirect printing of alerts from `level` to `file`."""
    cfg.fd[level] = file


def alert_level(level=None):
    """Get or set the verbosity level for the alert functions.

    err() will print something with level 0 (and greater), i.e. always.
    notice() will print something with level 1 (and greater).
    info() will print something with level 2 (and greater).
    debug() will print something with level 3 (and greater).
    trace() will print something with level 4 (and greater).
    """
    if level is not None:
        if type(level) is str:
            level = globals()[level]
        cfg.level = max(0, min(level, cfg.max_level))
    return cfg.level

def alcf():
    return cfg

def alert_level_name(level=None):
    """Return the name of the specified (or current) level number."""
    if level is None:
        level = cfg.level
    return alert_levels[level][0]


def alert_level_up():
    """Increase the alert level by one.

    This is intended to be used as the callback function for the default value
    of a pgetopt option to increase the verbosity.

    """
    if cfg.level < cfg.max_level:
        cfg.level += 1
    return cfg.level


def alert_level_zero():
    """Set the alert level to zero (errors only).

    This is intended to be used as the callback function for the default value
    of a pgetopt option to set the verbosity to zero.

    """
    cfg.level = 0
    return cfg.level


def is_notice():
    """Return True iff the alert level is at least at notice."""
    return cfg.level >= L_NOTICE

def is_info():
    """Return True iff the alert level is at least at info."""
    return cfg.level >= L_INFO

def is_debug():
    """Return True iff the alert level is at least at debugging."""
    return cfg.level >= L_DEBUG

def is_trace():
    """Return True iff the alert level is at least at tracing."""
    return cfg.level >= L_TRACE


def alert_if_level(level, *msgs):
    """Print a message if `level` is <= the cfg.level.

    If a decoration exists in `cfg.decoration[]` for that level, is it prepended
    to the message. By default, all levels print to stderr; this can be changed
    in `cfg.fd[]` by level.

    If one of the elements in `msgs` is a callable, it will be called withaout
    arguments to get the value of the element. This way, compute-intensive task
    can be delayed to the alerting moment, meaning they don't need to be done if
    not called for.

    """
    # return fast if not needed
    if level > cfg.level:
        return

    # make all msgs elements strings, calling those that are callable
    msgs = list(msgs)                   # is a tuple before
    for i, elem in enumerate(msgs):
        if callable(elem):
            msgs[i] = elem()
        else:
            msgs[i] = str(elem)
    if cfg.decoration[level]:
        msgs = [cfg.decoration[level].format(**globals()), *msgs]
    if cfg.timestamps:
        msgs.insert(0, cfg.timestamps())

    channel = cfg.fd[level]
    channel = { 1: sys.stdout, 2: sys.stderr }.get(channel) or channel

    msgtext = " ".join(msgs).rstrip()
    print(msgtext, file=channel, flush=True)

    if cfg.syslog_facility and cfg.syslog_prio[level]:
        if not cfg.syslog_opened:
            syslog.openlog(logoption=syslog.LOG_PID,
                           facility=cfg.syslog_facility)
            cfg.syslog_opened = True
        level = max(0, min(cfg.max_level, level))
        message = " ".join(map(str, msgs))
        syslog.syslog(cfg.syslog_prio[level], message)


def debug_vars(*vars):
    """Print debug output for the named variables if alert level >= L_DEBUG."""
    if cfg.level >= L_DEBUG:
        context = inspect.currentframe().f_back.f_locals
        for var in vars:
            debug("VAR {}: {}".format(var, repr(context[var])))


def err(*msgs):
    """Print error level output."""
    cfg.had_errors = True
    alert_if_level(L_ERROR, *msgs)
error = err                             # alias

def fatal(*msgs, exit_status=1):
    """Print error level output."""
    alert_if_level(L_ERROR, "Fatal", *msgs)
    sys.exit(exit_status)

def notice(*msgs):
    """Print notice level output according to alert level."""
    alert_if_level(L_NOTICE, *msgs)

def info(*msgs):
    """Print info level output according to alert level."""
    alert_if_level(L_INFO, *msgs)

def debug(*msgs):
    """Print debug level output according to alert level."""
    alert_if_level(L_DEBUG, *msgs)
dbg = debug                             # alias

def trace(*msgs):
    """Print debug level output according to alert level."""
    alert_if_level(L_TRACE, *msgs)

# EOF



# The StrBuf is used by the parser that interprets a parameters command.

class StringReader():
    """A class that lets the user read characters from the underlying string.

    Alas, io.StringIO doesn't have a straightforward eof() method, which would
    have made this implementation unnecessary.

    """
    def __init__(self, content):
        """Create a strBuf from a string."""
        self.content = content
        self.len = len(content)
        self._next = 0

    def next(self):
        """Return the next character from the strBuf, or None at EOF."""
        if self._next < self.len:
            ch = self.content[self._next]
            self._next += 1
            return ch

    def backup(self):
        if self._next > 1:
            self._next -= 1

    def eof(self):
        """Return True if the strBuf is at EOF."""
        return self._next >= self.len

    def __str__(self):
        """Return a string representation of a strBuf.

        The next character is marked by a preceding '^'; this will of course be
        not very helpful with a string containing this character.

        """
        head = self.content[:self._next]
        tail = self.content[self._next:]
        show = head + "^" + tail
        return "<{} {}>".format(self.__class__.__name__, repr(show))




#!/usr/bin/env python3

from jpylib import is_trace, trace

def tracefn(func):
    """Decorator: trace decorated function's calls if trace level is set."""
    def wrapper(*args, **kwargs):
        if is_trace():
            s = "call {}({}".format(func.__name__, ', '.join(map(repr, args)))
            if kwargs:
                for k, v in kwargs.items():
                    s += ", {}={}".format(k, repr(v))
            trace(s + ")")
        return func(*args, **kwargs)
    return wrapper




# time-related functions
# (I can never remmeber them, so I'll have my own.)

from datetime import datetime


def time_format(format, time):
    if time is None:
        time = datetime.now()
    return time.strftime(format)


def isotime(time=None):
    return time_format("%Y%m%d:%H%M%S", time)


def iso_time(time=None, sep="T"):
    return time_format("%Y-%m-%d{}%H:%M:%S".format(sep), time)


def iso_time_us(time=None, sep="T"):
    return time_format("%Y-%m-%d{}%H:%M:%S.%f".format(sep), time)


def iso_time_ms(time=None, sep="T"):
    return time_format("%Y-%m-%d{}%H:%M:%S.%f".format(sep), time)[:-3]




import os
import sys
import setuptools

repository = "https://git.w21.org/ni/jpylib/"
pkg_version = os.environ["PKG_VERSION"] or sys.exit("missing $PKG_VERSION env.")

with open("README.md", "r") as fh:
    # filter out the comment lines from the README
    description_lines = [
        l.replace("$__package_version$", pkg_version)
        for l in fh if not l.startswith("#")
    ]

setuptools.setup(
    name="jpylib-jyrgenn",
    version=pkg_version,
    author="Juergen Nickelsen",
    author_email="ni@w21.org",
    description=description_lines[0].strip(),
    long_description="".join(description_lines),
    long_description_content_type="text/markdown",
    url=repository,
    packages=setuptools.find_packages(),
    project_urls=dict(
        Documentation="{}/-/blob/v{}/README.md".format(repository, pkg_version),
        Source=repository,
    ),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.5",
)



#!python3

"""This module implements two context managers. outputCaptured()
captures output to stdout and stderr from the code run in the
context; in addition, outputAndExitCaptured() also captures a
sys.exit() call from the code run in the context and the status
value passed to sys.exit().

The latter will fail if the code run in the context catches
_SysExitException, which can be the case if the code catches all
of class Exception.
"""

# The meat of the first part is from Rob Kennedy on stackoverflow:
# https://stackoverflow.com/questions/4219717/how-to-assert-output-with-nosetest-unittest-in-python

import sys
from io import StringIO
from contextlib import contextmanager

@contextmanager
def outputCaptured():
    """Context manager to capure output to stdout and stderr.

    This works by temporarily replacing sys.stdout and sys.stderr with
    StringIO ports; these are both returned, so the output of the code
    run on the context can be retrieved from them:

        with outputCaptured() as (out, err):
            ...
        theOutput = out.getvalue()        # stdout output as string
        theErrout = err.getvalue()        # stderr outout as string
    """
    saved_out = sys.stdout
    saved_err = sys.stderr
    try:
        sys.stdout = StringIO()
        sys.stderr = StringIO()
        
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout = saved_out
        sys.stderr = saved_err


# From here, it's my own, built on the above. I feel pretty smug not only
# capturing the sys.exit(), but also the exit status value at the first
# attempt. ¡yay Python!

class _SysExitException(Exception):
    """Exception to be used by the fake sys.exit() function."""
    
    def __init__(self, status):
        """Keep the status value for later perusal."""
        self.status = status

        
class _ExitStatus():
    """Bog standard namespace. We use the `value` field.

    To keep the exit status, we need to return something from the context
    manager where we can store the exit status later, and this is it.
    """
    value = None


def _fake_sysexit(status=0):
    """Replacement for sys.exit() in the capturedOutputExit context manager.

    It throws the control flow out of code's context by raising an
    exception. (This will of course fail if the code run in the context
    captures all exceptions; let's just say it shouldn't.) The exit
    status value is passed to the context manager as the exception
    parameter.

    """
    raise _SysExitException(status)


@contextmanager
def outputAndExitCaptured():
    """Context manager to capture output to stdout/stderr and exit status.

    Like with outputCaptured(), stdout and stderr are captured in the
    returned StringIO objects. In addition, the exit status in case of a
    sys.exit() is captured in the `value` property of the returned
    status object:

        with outputAndExitCaptured() as (out, err, status):
            ...
        theOutput = out.getvalue()    # stdout output as string
        theErrout = err.getvalue()    # stderr outout as string
        theStatus = status.value      # sys.exit() argument (or 0 or None)

    The status value is None if the code run in the context hasn't
    called sys.exit().

    """
    saved_out = sys.stdout
    saved_err = sys.stderr
    saved_sysexit = sys.exit
    exit_status = _ExitStatus()
    
    try:
        sys.stdout = StringIO()
        sys.stderr = StringIO()
        sys.exit = _fake_sysexit
        
        yield sys.stdout, sys.stderr, exit_status
    except _SysExitException as syse:
        exit_status.value = syse.status
    finally:
        sys.stdout = saved_out
        sys.stderr = saved_err
        sys.exit = saved_sysexit


# EOF



# parser for key=value strings

"""Parser for key-value strings like "foo=bar,dang=[1,2,15],d={a=b,c=[d,e,f]}".

The data is returned as a Python data structure composed of strings,
dictionaries, and lists. It is used to set configuration values from
command-line arguments with a syntax more compact than e.g. JSON.

Syntax:

 * On the top level, the key-value string is a kvpairs list.

 * A kvpairs list is a list of zero or more key=value pairs separated by commas.
   It is mapped to a Python dictionary. Example:
   "signals=[1,2,15],action=terminate"

 * A key is a string that does not contain curly brackes, brackets, a comma, or
   an equals sign. Leading whitespace is not considered part of the key;
   trailing or embedded whitespace is a syntax error. For configuration values,
   it is helpful to match the syntax of Python identifiers, i.e. first character
   an underscore or a letter, following characters, if any, underscore, letter,
   or digit. These are mapped to Python dictionary keys. Example: "key_file"

 * A value can be a literal, a dictionary, or a list of values.

 * A literal value is a string of characters that doesn't contain curly brackes,
   brackets, a comma, or an equals sign. Whitespace is considered part of the
   literal. These are mapped to Python strings. Example: "Radio Dos"

 * A dictionary is a kvpairs list enclosed by curly braces. Example:
   "{file=~/etc/foo.conf,syntax=INI}"

 * A list is a list of zero or more values separated by commas and enclosed in
   brackets. Example: "[HUP,INTR,TERM]"

This syntax is obviously limited, but sufficient to express complex data
structures with (some) string values as leaves. It is mainly meant to be compact
for use on the command line.

The parser is somewhat sloppy and will accept some deviations from this
descriptsion, but exploiting this sloppyness will not be of any use.

"""

from .stringreader import StringReader

# def debug(*msg):
#     print("DBG", *msg)

class SyntaxError(Exception):
    """An exception raised when the parser sees a syntax error.

    Its string argument is a string representation of the underlying
    StringReader with a marker behind the letter where the error was seen.

    """
    pass

separators = "{}[],="


def syntax_error(buf, message, *args):
    raise SyntaxError(str(buf), message.format(*args) if args else message)


def next_token(buf, intvals=False):
    token = ""
    while not buf.eof():
        ch = buf.next()
        if ch in separators:
            if token:
                buf.backup()
            else:
                token = ch
            break
        token += ch
    if intvals:
        try:
            token = int(token)
        except:
            pass
    # debug("next_token", repr(token))
    return token or None


def parse_valuelist(buf, intvals=False):
    """Parse a list of values."""
    result = []
    while True:
        # state: want a value
        t = next_token(buf, intvals=intvals)
        if t is None:
            syntax_error(str(buf), "value list misses closing ']'")
        elif t == "]":
            if result:
                syntax_error(str(buf), "value list misses value after ','")
            return result
        elif t == "[":
            result.append(parse_valuelist(buf))
        elif t == "{":
            result.append(parse_kvpairs(buf, need_brace=True, intvals=intvals))
        elif str(t) in "=}":
            syntax_error(buf, "unexpected in {} value list", repr(t))
        elif t == ",":
            result.append("")
            continue            # don't wait for comma, had one already
        else:
            result.append(t)
        # now want a comma
        t = next_token(buf, intvals=intvals)
        if t is None:
            syntax_error(str(buf), "value list misses closing ']'")
        elif t == "]":
            return result
        elif t in "[{}=":
            syntax_error(buf, "unexpected in {} value list", repr(t))
        elif t == ",":
            pass                # *so* pass!
        # nothing else can happen here -- I hope


def parse_kvpairs(buf, need_brace=False, intvals=False):
    """Parse a kvpairs list `key=value,...`."""
    result = {}
    while True:
        # read key
        t = next_token(buf, intvals=intvals)
        if t is None:
            if need_brace:
                syntax_error(str(buf), "kvpairs list misses closing '}'")
            return result
        if t == "}":
            if not need_brace:
                syntax_error(str(buf), "unexpected '}' in top-level kvpairs")
            return result
        if t == ",":
            continue
        if str(t) in separators:
            syntax_error(str(buf), "unexpected '{}' in kvpairs list", t)
        key = t
        # expect "="
        t = next_token(buf, intvals=intvals)
        if t != "=":
            syntax_error(str(buf), "expected '=' after key in kvpairs list")
        # read value
        t = next_token(buf, intvals=intvals)
        if t == "{":
            result[key] = parse_kvpairs(buf, need_brace=True, intvals=intvals)
            continue
        if t == "[":
            result[key] = parse_valuelist(buf, intvals=intvals)
            continue
        if t == ",":
            result[key] = ""
            continue
        if t is None:
            if need_brace:
                syntax_error(str(buf), "kvpairs list misses closing '}'")
            result[key] = ""
            return result
        result[key] = t


def parse_kvs(string, intvals=False):
    """Parse a key=value string and return the data structure."""
    return parse_kvpairs(StringReader(string), intvals=intvals)



#!/usr/bin/env python3

class Namespace:
    """Simple name space class as a key-value store.

    Values can be assigned and read directly (ns.key = value) or using the
    set()/get() methods. An update() method (as with a dictionary) has the
    options to skip keys beginning with an underscore, or raise a KeyError if a
    key is not previously known.

    """

    def __init__(self, **kwargs):
        """Initialize a Namespace object from the 'kwargs' mapping."""
        self.__dict__.update(kwargs)

    def update(self, new_values, skip_underscore=False, reject_unknown=False):
        """Update the object with a dictionary of new key/value pairs.

        If `reject_unknown` is true, it is an error if the argument dictionary
        contains keys that are not in the object's key set.

        If `skip_underscore` is true, keys that start with an underscore ("_")
        are not considered for update.

        """
        for key, value in new_values.items():
            if key.startswith("_") and skip_underscore:
                continue
            if reject_unknown and key not in self.__dict__:
                raise KeyError("unknown key in {}: {}".format(
                    self.__class__.__name__, key))
            self.__dict__[key] = value

    def set(self, key, value):
        """Set a value."""
        self.__dict__[key] = value

    def get(self, key, default=None):
        """Get a value; default if key is not present."""
        return self.__dict__.get(key, default)

    def __str__(self):
        """Return a string repr in the form of '<class>(key1=value1, ...)'."""
        return self.__class__.__name__ + "(" + ", ".join(
            ["{}={}".format(k, repr(v)) for k, v in self.__dict__.items()]) \
            + ")"

    def __repr__(self):
        return self.__str__()
